# flake8: noqa
import importlib.metadata

from ixmp4.core import (
    Platform as Platform,
    Region as Region,
    Unit as Unit,
    Run as Run,
    Variable as Variable,
    Model as Model,
    Scenario as Scenario,
)

from ixmp4.core.exceptions import (
    NotFound as NotFound,
    NotUnique as NotUnique,
    IxmpError as IxmpError,
    InconsistentIamcType as InconsistentIamcType,
)

from ixmp4.data.abstract import DataPoint as DataPoint

__version__ = importlib.metadata.version("ixmp4")
